#if 0
 gcc -Wall -Os -static -s -o karl karl.c
 exit 0 
#endif

/********************************************************************\
* KARL - Keyboard Autonomous Reflexive Learning                      *
* Learn keyboard keys meaning by touch.                              *
* (C) Klaus Knopper 2013                                             *
* License: GPL V2                                                    *
* Code greatly inspired from showkey.c from console-tools package.   *
\********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <memory.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <sysexits.h>
#include <ctype.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <locale.h> /* LC_ALL */
#include <errno.h>

/* Debug */
int verbose=0;

#define LOCALEDIR "/usr/share/locale"

# include <libintl.h>
# define _(String) gettext (String)
# ifdef gettext_noop
#  define N_(String) gettext_noop (String)
# else
#  define N_(String) (String)
# endif

/* setup localization for a program */
#define setuplocale() do { \
  setlocale (LC_ALL, ""); \
  bindtextdomain ("karl", LOCALEDIR); \
  textdomain ("karl"); \
} while (0)


int is_a_console(int fd) 
{
  char arg = 0;
  return (ioctl(fd, KDGKBTYPE, &arg) == 0
	  && ((arg == KB_101) || (arg == KB_84)));
}

static int open_a_console(char *fnam) 
{
  int fd;
  
  /* try read-write */
  fd = open(fnam, O_RDWR);
  
  /* on failure, try read-only */
  if (fd < 0 && errno == EACCES)
      fd = open(fnam, O_RDONLY);
  
  /* on failure, try write-only */
  if (fd < 0 && errno == EACCES)
      fd = open(fnam, O_WRONLY);
  
  /* on failure, fail */
  if (fd < 0)
      return -1;
  
  /* if not a console, fail */
  if (! is_a_console(fd))
    {
      close(fd);
      return -1;
    }
  
  /* success */
  return fd;
}

/* Turn echo on or off on console */
void echo(int d, int on)
{
 struct termios t;
 if(ioctl(d,TCGETS,&t)>=0)
  {
   if(on) t.c_lflag|=ECHO; else t.c_lflag&=~ECHO;
   ioctl(d,TCSETS,&t);
  }
}


/*
 * Get an fd for use with kbd/console ioctls.
 * Some ioctls require root, so this program is supposed to run via suid.
 *
 * if tty_name is non-NULL, try this one instead.
 */

int get_console_fd(char* tty_name) 
{
  int fd;

  if (tty_name)
    {
      if (-1 == (fd = open_a_console(tty_name)))
	return -1;
      else
	return fd;
    }
  
  fd = open_a_console("/dev/tty");
  if (fd >= 0)
    return fd;
  
  fd = open_a_console("/dev/tty0");
  if (fd >= 0)
    return fd;
  
  fd = open_a_console("/dev/console");
  if (fd >= 0)
    return fd;
  
  for (fd = 0; fd < 3; fd++)
    if (is_a_console(fd))
      return fd;
  
  fprintf(stderr,
	  _("Couldn't get a file descriptor referring to the console\n"));
  return -1;		/* total failure */
}


int fd;
int oldkbmode;
struct termios old, old_term;
struct kbd_repeat kbdrep_old = { -1, -1 };

void get_mode(void)
{

  if (ioctl(fd, KDGKBMODE, &oldkbmode))
    {
      perror("ioctl(KDGKBMODE)");
      exit(1);
    }
  if (ioctl( fd, KDKBDREP, &kbdrep_old ))
    {
	perror("ioctl(KDKBDREP)");
	exit( 1 );
    }
}

void clean_up(void)
{
  if (ioctl(fd, KDSKBMODE, oldkbmode))
    {
      perror("ioctl(KDSKBMODE)");
      exit(1);
    }
  if (tcsetattr(fd, 0, &old) == -1)
    perror("tcsetattr");

  if (tcsetattr(0, TCSAFLUSH, &old_term) == -1)
    perror("tcsetattr = %d\n");

  if (ioctl( fd, KDKBDREP, &kbdrep_old ))
    {
      perror( "ioctl(KDKBDREP)" );
    }

  echo(0,1);

  close(fd);
}

void die(int x)
{
  printf(_("caught signal %d, cleaning up...\n"), x);
  clean_up();
  exit(1);
}

int waitforkey(int d, int seconds)
{
 struct timeval tv = { seconds, 0L };
 fd_set rfds;
 FD_ZERO(&rfds);
 FD_SET(d, &rfds);
 return select(fd+1, &rfds, NULL, NULL, &tv);
}

int main (int argc, char *argv[])
{
  struct termios new;
  struct kbd_repeat kbdrep_s = { 0, 0 };
  struct kbentry kb;
  unsigned char buf[129];
  int i, n, skip=0;

  setuplocale();

  if (-1 == (fd = get_console_fd(NULL))) exit (1);

  signal(SIGHUP, die);  signal(SIGINT, die);  signal(SIGQUIT, die); signal(SIGILL, die);
  signal(SIGTRAP, die); signal(SIGABRT, die); signal(SIGIOT, die);  signal(SIGFPE, die);
  signal(SIGKILL, die); signal(SIGUSR1, die); signal(SIGSEGV, die); signal(SIGUSR2, die);
  signal(SIGPIPE, die); signal(SIGTERM, die); signal(SIGCHLD, die); signal(SIGCONT, die);
  signal(SIGSTOP, die); signal(SIGTSTP, die); signal(SIGTTIN, die); signal(SIGTTOU, die); 
#ifdef SIGSTKFLT
  signal(SIGSTKFLT, die);
#endif

  get_mode();

  if (tcgetattr(fd, &old) == -1)
    perror("tcgetattr = %d\n");
  if (tcgetattr(0, &old_term) == -1)
    perror("tcgetattr = %d\n");
  if (tcgetattr(fd, &new) == -1)
    perror("tcgetattr = %d\n");

  /* new.c_lflag &= ~ (ICANON | ECHO | ISIG); */
  new.c_lflag &= ~ (ICANON | ECHO | ISIG);
  new.c_iflag = 0;
  new.c_cc[VMIN] = sizeof(buf);
  new.c_cc[VTIME] = 1;	/* 0.1 sec intercharacter timeout */

  /* K_MEDIUMRAW K_XLATE K_RAW))) */
  if (ioctl(fd, KDSKBMODE, K_RAW ))
    {
      perror("KDSKBMODE");
      exit(1);
    }

  if (ioctl( fd, KDKBDREP, &kbdrep_s ))
    {
      perror( "ioctl(KDKBDREP)" );
      exit( 1 );
    }

  if (tcsetattr(fd, TCSAFLUSH, &new) == -1)
    perror("tcsetattr = %d\n");

  if (tcsetattr(0, TCSAFLUSH, &new) == -1)
    perror("tcsetattr = %d\n");

  echo(0,0);

  /* Let screenreader read the help and wait for text, but skip to key code */
  /* evaluation if a key was pressed. */
  {
   struct { char *text; int delay; } help[3] =
    { { _("KARL - Keyboard Autonomous Reflexive Learning"), 4},
      { _("Type any keys to learn their meaning."), 5},
      { _("Hit Escape key on upper left to exit program."), 5 } };
   for(i=0; i<3; i++)
    {
     printf("\n\r %s\r", help[i].text); fflush(stdout);
     if(skip) continue;
     if(waitforkey(fd, help[i].delay)) skip=1;
    }
   printf("\n\r"); fflush(stdout);
  }

  while (1) {
   unsigned int code=0;
   memset(buf,0,sizeof(buf));
   n = read(fd, buf, 1);
   if(n>0){
    if(buf[0] & 0x80) continue;
    printf("\r\033[K\r "); fflush(stdout); usleep(125000);
    if(verbose) printf("%d-byte code:", n);
    for(i=0; i<n;i++){
     code = buf[i]; 
     if(verbose) printf(" 0x%02x", code);
    }
    kb.kb_table=0;
    kb.kb_index= code & (NR_KEYS-1);
    if(!ioctl(fd, KDGKBENT, &kb)){
     int literal = kb.kb_value & 0xff;
     char *special = NULL;
     if(verbose) printf(" -> %d -> ", kb.kb_value);
     switch(kb.kb_value){
      case 9:    special=_("Tabulator key"); break;
      case 27:   printf("\r\033[1m %s\033[0m\033[K", _("Escape - program exit")); fflush(stdout); sleep(3); special=""; break;
      case 32:   special=_("Space key"); break;
      case 127:  special=_("Backspace erase to left"); break;
      case 223:  special=_("Minus or Underscore"); break;
      case 256:  special=_("F1"); break;
      case 257:  special=_("F2"); break;
      case 258:  special=_("F3"); break;
      case 259:  special=_("F4"); break;
      case 260:  special=_("F5"); break;
      case 261:  special=_("F6"); break;
      case 262:  special=_("F7"); break;
      case 263:  special=_("F8"); break;
      case 264:  special=_("F9"); break;
      case 265:  special=_("F10"); break;
      case 266:  special=_("F11"); break;
      case 267:  special=_("F12"); break;
      case 512:  /* Special case! */
                 switch(code){
		  case 0x0c: special=_("Minus or Underscore"); break;
		  case 0x1a: special=_("Left square bracket or Left brace"); break;
		  case 0x27: special=_("Semicolon or Colon"); break;
		  case 0x28: special=_("single quote or double quote"); break;
		  case 0x5b: special=_("Super left"); break;
		  case 0x5c: special=_("Super right"); break;
		  case 0x5d: special=_("Menu right"); break;
		 }; break;
      case 513:  special=_("Enter"); break;
      case 519:  special=_("Caps Lock"); break;
      case 520:  special=_("Pause"); break;
      case 521:  special=_("Scroll Lock"); break;
      case 768:  special=_("Insert"); break;
      case 769:  special=_("End of page"); break;
      case 770:  special=_("Arrow down"); break;
      case 771:  special=_("Page down"); break;
      case 772:  special=_("Arrow left"); break;
      case 774:  special=_("Arrow right"); break;
      case 775:  special=_("Beginning of page"); break;
      case 776:  special=_("Arrow up"); break;
      case 777:  special=_("Page up"); break;
      case 780:  special=_("Print"); break;
      case 783: case 784:  special=_("Delete"); break;
      case 1026: special="^"; break;
      case 1792: special=_("Shift"); break;
      case 1794: special=_("Control"); break;
      case 1795: special=_("Alt"); break;
      /* German and french umlauts, may need translation, too */
      case 3040: special="à"; break;
      case 3044: special="ä"; break;
      case 3047: special="ç"; break;
      case 3048: special="è"; break;
      case 3049: special="é"; break;
      case 3062: special="ö"; break;
      case 3065: special="ö"; break;
      case 3068: special="ù"; break;
     }
     if(special) printf(special);
     else if(isprint(literal)){
      switch(literal){
       case '1': printf(_("1 or exclamation mark")); break;
       case '2': printf(_("2 or at sign")); break;
       case '3': printf(_("3 or hash mark")); break;
       case '4': printf(_("4 or dollar sign")); break;
       case '5': printf(_("5 or percent sign")); break;
       case '6': printf(_("6 or circumflex")); break;
       case '7': printf(_("7 or ampersand")); break;
       case '8': printf(_("8 or asterisk")); break;
       case '9': printf(_("9 or bracket left")); break;
       case '0': printf(_("0 or bracket right")); break;
       case '^': printf(_("Zirkumflex oder Gradzeichen")); break;
       case '`': case '\047': printf(_("Grave accent or Tilde")); break;
       case '#': printf(_("Raute oder einfaches Anführungszeichen Apostroph")); break;
       case '\\': printf(_("Backslash or Pipe")); break;
       case '=': printf(_("Equal sign or Plus")); break;
       case '+': printf(_("Plus oder Stern oder Tilde")); break;
       case '-': printf(_("minus or underscore")); break;
       case '.': printf(_("Punct or greater than sign")); break;
       case ',': printf(_("Comma or smaller than sign")); break;
       case '<': printf(_("greater than sign or smaller than sign")); break;
       case '[': printf(_("left square bracket or left brace")); break;
       case ']': printf(_("right square bracket or right brace")); break;
       case '/': printf(_("slash or question mark")); break;
       default:  printf("%c", literal); break;
      }
     } else {
      printf(_("Unknown key code %d"),kb.kb_value);
     }
    }
    printf("\033[K\r"); fflush(stdout);
   }
   if(code == 0x01) break; /* Escape */
  }
/* 	s = (buf[i] & 0x80) ? _("release") : _("press"); */
 clean_up();
 printf("\n");
 exit(0);
}
